"""
A framework and supporting libraries
for working with Flask, Connexion, and OpenAPI.

For usage, see :ref:`ligare-web`.

For examples and guides, see :ref:`examples-web-applications`.
"""

__version__: str = "0.7.0"
